import readyoda
import numpy as np

class Run(object):
    '''Information about a run: labels, where to find its data, etc.'''
    def __init__(self, label, nominalpath, pdfvarpaths=[], scalevarpaths=[], altpdfpaths=[]):
        '''NOTE: `pdfvarpaths` are only the variations of the nominal set, not alternative sets.'''
        super(Run, self).__init__()
        self.label = label
        self.nominalpath = nominalpath
        self.pdfvarpaths = pdfvarpaths
        self.scalevarpaths = scalevarpaths
        self.altpdfpaths = altpdfpaths
    
    
    
    def binedges(self, histogram_path):
        return readyoda.get_bin_edges(histogram_path, self.nominalpath)
    
    
    
    def nominal(self, histogram_path):
        return readyoda.get_curve(histogram_path, self.nominalpath, self.nominalpath)
    
    def statistical_uncertainty_up_down(self, histogram_path):
        relative = readyoda.get_relative_error(histogram_path, self.nominalpath)
        return self.nominal(histogram_path) * (1. + relative), self.nominal(histogram_path) * (1. - relative)
    
    
    def pdfvars(self, histogram_path):
        if not self.pdfvarpaths:
            print('Warning: no PDF variations available, returning nominal curve instead!')
            return self.nominal(histogram_path)
        #print self.pdfvarpaths
        return readyoda.get_curves(histogram_path, self.pdfvarpaths, self.nominalpath)
    
    
    
    def scalevars(self, histogram_path):
        if not self.scalevarpaths:
            print('Warning: no QCD scale variations available, returning nominal curve instead!')
            return self.nominal(histogram_path)
        return readyoda.get_curves(histogram_path, self.scalevarpaths + [self.nominalpath], self.nominalpath)
        
    
    
    def altpdf(self, histogram_path, index=0):
        if not self.altpdfpaths:
            print('Warning: no alternative PDFs available, returning nominal curve instead!')
            return self.nominal(histogram_path)
        try:
            altpdfpath = self.altpdfpaths[index]
            return readyoda.get_curve(histogram_path, altpdfpath, self.nominalpath)
        except IndexError:
            raise RuntimeError('You requested the curve for the {0}. alternative PDF, but only {} alternative PDFs are given.'.format(index+1, len(self.altpdfpaths)))
    
    
    
    def pdf_variation_up_down(self, histogram_path, clip_to_zero=True):
        pdfvars = self.pdfvars(histogram_path)
        nominal = self.nominal(histogram_path)
        #BINWIDTH = self.binedges(histogram_path)[1] - self.binedges(histogram_path)[0]
        #print pdfvars[:0] * BINWIDTH # / np.sum(nominal) * 100.
        if not self.pdfvarpaths:
            print('Warning: no PDF variations given, uncertainty band will have zero width.')
            return pdfvars, pdfvars # the band has zero width because no PDF variations were given
        if not len(self.pdfvarpaths) == 100:
            print('Warning: expecting NNPDF 3.0 with 100 PDF variations plus nominal, found {0}. If you use a different PDF set, you need to implement its uncertainty calculation. I will proceed, but using the NNPDF prescription for uncertainties, which may be incorrect for your set.'.format(len(self.pdfvarpaths)))
        shifts = np.sqrt(np.sum((pdfvars - nominal)**2, axis=0) / 100.)
        up = nominal + shifts
        #print shifts / nominal * 100.
        down = (nominal - shifts).clip(min=0) if clip_to_zero else nominal - shifts
        return up, down
    
    
    
    def scale_variation_up_down(self, histogram_path):
        scalevars = self.scalevars(histogram_path)
        if len(scalevars.shape) == 1: return scalevars, scalevars
        return np.max(scalevars, axis=0), np.min(scalevars, axis=0)



    # Make sure the histogram you pass contains all the generated events
    # Returns RELATIVE uncertainty IN PERCENT
    def integrated_yield_uncertainty_scale(self, histogram_path):
        nominal = np.sum(self.nominal(histogram_path)[::2])
        scalevars = [np.sum(var[::2]) for var in self.scalevars(histogram_path)]
        up = (max(scalevars) - nominal) /  nominal * 100.
        down = (min(scalevars) - nominal) /  nominal * 100.
        return up, down


    # Return the scale uncertainty of the inclusive integrated yield
    # or, if jets >= 0, the yield uncertainty in that exclusive jet bin.
    # Returns RELATIVE uncertainty IN PERCENT
    def yield_uncertainty_scale(self, exclusive_njet_histogram_path, jets=-1):
        if jets < 0:
            return self.integrated_yield_uncertainty_scale(exclusive_njet_histogram_path)
        nominal = self.nominal(exclusive_njet_histogram_path)
        upvar, downvar = self.scale_variation_up_down(exclusive_njet_histogram_path)
        up = (upvar[::2][jets] - nominal[::2][jets]) / nominal[::2][jets] * 100.
        down = (downvar[::2][jets] - nominal[::2][jets]) / nominal[::2][jets] * 100.
        return up, down        



    # Make sure the histogram you pass contains all the generated events
    # Returns RELATIVE uncertainty IN PERCENT
    def integrated_yield_uncertainty_pdf(self, histogram_path):
        nominal = np.sum(self.nominal(histogram_path)[::2])
        pdfvars = [np.sum(var[::2]) for var in self.pdfvars(histogram_path)]
        shifts = np.sqrt(np.sum((pdfvars - nominal)**2, axis=0) / 100.)
        up = shifts /  nominal * 100.
        down = -shifts / nominal * 100.
        return up, down



    # Return the PDF uncertainty of the inclusive integrated yield
    # or, if jets >= 0, the yield uncertainty in that exclusive jet bin.
    # Returns RELATIVE uncertainty IN PERCENT
    def yield_uncertainty_pdf(self, exclusive_njet_histogram_path, jets=-1):
        if jets < 0:
            return self.integrated_yield_uncertainty_pdf(exclusive_njet_histogram_path)
        nominal = self.nominal(exclusive_njet_histogram_path)
        upvar, downvar = self.pdf_variation_up_down(exclusive_njet_histogram_path)
        up = (upvar[::2][jets] - nominal[::2][jets]) / nominal[::2][jets] * 100.
        down = (downvar[::2][jets] - nominal[::2][jets]) / nominal[::2][jets] * 100.
        return up, down





# Auxiliary modules
import numpy as np
from numpy.testing import assert_almost_equal as assert_approx
from os.path import join, dirname
import ROOT
import sys

# Module to be tested
sys.path.append(join(dirname(__file__), '..'))
import heppy


#
# Set up
#
testfile = join(dirname(__file__), 'example.root')
hist1d_path = 'root_histogram_1d_name'
hist2d_path = 'root_histogram_2d_name'
areas2d = np.array([[1., 5.],
       				[2., 6.],
       				[3., 7.],
       				[4., 8.]])
# Reference 2D ROOT histogram
reference_th2 = ROOT.TH2F('reference_th2', 'reference_th2', 4, -1.0, 1.0, 2, -1.0, 1.0)
# The zero padding is for the under-/overflows
reference_areas = np.transpose(np.array([
	[0., 0., 0., 0., 0., 0.],
	[0., 1., 2., 3., 4., 0.],
	[0., 5., 6., 7., 8., 0.],
	[0., 0., 0., 0., 0., 0.],]))
reference_area_errors = np.transpose(np.array([
	[0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
	[0.0, 0.4, 0.3, 0.2, 0.1, 0.0],
	[0.0, 0.8, 0.7, 0.6, 0.5, 0.0],
	[0.0, 0.0, 0.0, 0.0, 0.0, 0.0],]))
reference_th2.SetBinContent(1, 1, 1.0)
reference_th2.SetBinContent(2, 1, 2.0)
reference_th2.SetBinContent(3, 1, 3.0)
reference_th2.SetBinContent(4, 1, 4.0)
reference_th2.SetBinContent(1, 2, 5.0)
reference_th2.SetBinContent(2, 2, 6.0)
reference_th2.SetBinContent(3, 2, 7.0)
reference_th2.SetBinContent(4, 2, 8.0)
reference_th2.SetBinError(1, 1, 0.4)
reference_th2.SetBinError(2, 1, 0.3)
reference_th2.SetBinError(3, 1, 0.2)
reference_th2.SetBinError(4, 1, 0.1)
reference_th2.SetBinError(1, 2, 0.8)
reference_th2.SetBinError(2, 2, 0.7)
reference_th2.SetBinError(3, 2, 0.6)
reference_th2.SetBinError(4, 2, 0.5)


def test_read_histogram_1d():
	'''
	Test reading a ROOT TH1F and creating a Heppy histogram from it
	'''
	h = heppy.readroot(testfile, hist1d_path)
	assert isinstance(h, heppy.histogram1d)
	np.testing.assert_array_almost_equal( h.areas, [1.1, 4.3, 9.5, 16.7] )
	np.testing.assert_array_almost_equal( h.binsizes, [0.5, 0.5, 0.5, 0.5] )
	np.testing.assert_array_almost_equal( h.heights, h.areas / 0.5 )
	np.testing.assert_array_almost_equal( h.uncorr_variations['Statistical__1up'], [1.3, 4.9, 10.5, 18.1] )
	np.testing.assert_array_almost_equal( h.uncorr_variations['Statistical__1down'], [0.9, 3.7, 8.5, 15.3] )


def test_read_histogram_2d():
	'''
	Test reading a ROOT TH2F and creating a Heppy histogram from it
	'''
	h = heppy.readroot(testfile, hist2d_path)
	assert isinstance(h, heppy.histogram2d)
	np.testing.assert_array_almost_equal( h.areas, areas2d )
	np.testing.assert_array_almost_equal( h.binsizes, np.ones_like(areas2d) * 0.5 )
	np.testing.assert_array_almost_equal( h.heights, areas2d / 0.5 )
	statistical__1up = (reference_areas + reference_area_errors)[1:-1,1:-1]
	np.testing.assert_array_almost_equal( h.uncorr_variations['Statistical__1up'], statistical__1up )
	statistical__1down = (reference_areas - reference_area_errors)[1:-1,1:-1]
	np.testing.assert_array_almost_equal( h.uncorr_variations['Statistical__1down'], statistical__1down )



# Generator of all paths to non-directories in the file
# Copied from Andy Buckley, http://pastebin.com/GebcyHY9
def path_generator(d, basepath='/'):
    for key in d.GetListOfKeys():
        kname = key.GetName()
        if key.IsFolder() and 'TDirectory' in key.GetClassName():
            for i in path_generator(d.Get(kname), basepath+kname+'/'):
                yield i
        else:
            yield basepath + kname #, d.Get(kname)



def string_contains_all(string, all):
    for a in all:
        if not a in string:
            return False
    return True



def string_contains_some(string, some):
    if not some:
        return True
    for s in some:
        if s in string:
            return True
    return False



def string_contains_none(string, none):
    for n in none:
        if n in string:
            return False
    return True



# Get list of all histogram paths in a ROOT file that pass filters:
# @param include     list of strings that paths need to contain to be considered
# @param exclude     list of strings that paths may not contain to be considered  
def get_keys(fpath, all=[], some=[], none=[]):
    import ROOT
    ROOT.PyConfig.IgnoreCommandLineOptions = True
    rootfile = ROOT.TFile(fpath, 'r')
    rootfile.cd()
    paths = []
    for path in path_generator(rootfile):
        if string_contains_all(path, all) and string_contains_some(path, some) and string_contains_none(path, none):
            paths.append(path)
    return paths



import ROOT
from array import array

def _array2hist_1d(nbins, areas, th1, errors=None):
    for bin_index in range(nbins):
        th1.SetBinContent(bin_index+1, areas[bin_index])
        if errors is not None:
            th1.SetBinError(bin_index+1, errors[bin_index])

def to_th1(name, binedges, areas, nbins, errors=None):
    th1 = ROOT.TH1D(name, name, len(binedges)-1, array('d', [b for b in binedges]))
    _array2hist_1d(nbins, areas, th1, errors=errors)
    return th1

def _array2hist_2d(nbins, areas, th2, errors=None):
    nbins_x, nbins_y = nbins
    for bin_index_x in range(nbins_x):
        for bin_index_y in range(nbins_y):
            th2.SetBinContent(bin_index_x+1, bin_index_y+1, areas[bin_index_x, bin_index_y])
            if errors is not None:
                th2.SetBinError(bin_index_x+1, bin_index_y+1, errors[bin_index_x, bin_index_y])

def to_th2(name, binedges, areas, nbins, errors=None):
    th2 = ROOT.TH2D(name, name, len(binedges[0])-1, array('d', [b for b in binedges[0]]), len(binedges[1])-1, array('d', [b for b in binedges[1]]))
    _array2hist_2d(nbins, areas, th2, errors=errors)
    return th2



from heppy.value import Value
from heppy.histogram import *
from heppy.readroot import * # uses uproot
import heppy.readroot_legacy # uses ROOT and root-numpy
from heppy.readtrex import readtrex
from heppy.panel import *
from heppy.plot import *
from heppy.heatmap import *
import heppy.scanroot # uses ROOT
import heppy.uncertainty
